from contextlib import suppress

from datetime import datetime, timedelta


def process_response(response: dict) -> dict:
    if response.get("STATUS"):
        match response.get("STATUS"):
            case 'S': 
                response['STATUS'] = 'success'
            case 'E':
                response['STATUS'] = 'error'
            case _:
                response['STATUS'] = None
    
    for key, value in response.items():
        if type(value) == dict:
            response[key] = process_response(value)
        elif type(value) == str:
            response[key] = value.strip()
            
            match value:
                case 'true':
                    response[key] = True
                case 'false':
                    response[key] = False
                case 'enable':
                    response[key] = True
                case 'disable':
                    response[key] = False
                case 'Alive':
                    response[key] = True
                case 'Dead':
                    response[key] = False
                case 'Y':
                    response[key] = True
                case 'N':
                    response[key] = False
                case '':
                    response[key] = None
                    
            if type(response[key]) == str and response[key].isdigit():
                response[key] = int(response[key])  
            with suppress(ValueError, TypeError):
                response[key] = float(response[key]) if not isinstance(response[key], int) else response[key]
        
        match key:
            case 'When':
                response[key] = datetime.fromtimestamp(response[key])
            case 'Uptime':
                response[key] = timedelta(seconds=response[key])
            case 'Elapsed':
                response[key] = timedelta(seconds=response[key])
            case 'Upfreq Complete':
                response[key] = bool(response[key])
            case 'upfreq_complete':
                response[key] = bool(response[key])
            case 'enable':
                response[key] = bool(response[key])
                
    return response


from pywhatsminer.core import WhatsminerAccessToken, WhatsminerAPI
from pywhatsminer.core.methods.power import Power
from pywhatsminer.core.methods.system import System
from pywhatsminer.core.methods.config import Config


class Client:
    def __init__(self, ip: str, port: int = 4028, password: str | None = "admin"):
        self._access_token = WhatsminerAccessToken(ip, port, password)
        self.api = WhatsminerAPI()

        self.Power = Power(self)
        self.System = System(self)
        self.Config = Config(self)


from pywhatsminer.core.api import WhatsminerAPI, WhatsminerAccessToken
from pywhatsminer.core.models import PSU
from pywhatsminer.core.utils import process_response

from typing import Any


class Power:
    def __init__(self, client):
        self.client = client
        self.api: WhatsminerAPI = client.api
        self.token: WhatsminerAccessToken = client._access_token
        
    
    def on(self) -> Any:
        """
        This operation simply starts mining and turns on the power output of the power board.
        """
        return self.api.exec_command(self.token, "power_on")
    

    def off(self, respbefore: bool = False) -> Any:
        """
        This operation simply stops mining and turns off the power output of the power board.
        There was no power outage on the control board
        
        Params:
            - respbefore: bool = False - when true, the miner returns the data first and then closes the hashboard, otherwise close the hashboard first and then return the data.
        """
        return self.api.exec_command(self.token, "power_off", {"respbefore": respbefore})
        
    
    def restart(self) -> Any:
        """
        This operation restarts the miner.
        """
        return self.api.exec_command(self.token, "restart_btminer")
    
    
    def switch_mode(self, mode: str) -> Any:
        """
        This operation changes miner's power mode.
        
        Params:
            - mode: str - new power mode. Possible fields: "low", "normal", "high"
        """
        assert mode in ["low", "normal", "high"]
        
        return self.api.exec_command(self.token, f"set_{mode}_power")


    def get_psu(self) -> PSU:
        """
        This method returns miner's power system status.
        """
        data = process_response(self.api.exec_command(self.token, "get_psu"))

        return PSU(*data['Msg'].values())
    
    
    def set_percent(self, percent: int) -> Any:
        """
        Temporarily set the power percent based on the current running power. 
        It is recommended to be used for temporary adjustment. Long run maybe unstable. 
        If you want the machine run for a long time in low power percent, see `Client.Power.adjust_limit()`
        
        Params:
            - percent: int - new power output percentage. Range: 0 to 100.
        """
        return self.api.exec_command(self.token, "set_power_pct", {"percent": percent})
    
    
    def adjust_limit(self, limit: int) -> Any:
        """
        This operation sets the upper limit of the miner's power. 
        Not higher than the ordinary power of the machine.
        If the Settings take effect, the machine will restart.

        Params:
            - limit: int - new power limit percentage. Range: 0 to 99999.
        """
        assert limit in range(0, 100000)
        
        return self.api.exec_command(self.token, "adjust_power_limit", {"power_limit":limit})
    
    
    def pre_power_on(self) -> Any:
        """
        The miner can be preheated by `pre_power_on` before `power_on`, so that the machine
        can quickly enter the full power state when "power on" is used. 
        
        You can also use this
        command to query the pre power on status. Make sure `power_off` btminer before
        `pre_power_on`. 
        
        Response:
            - complete: bool
            - query: str - query the status. Possible fields: "wait for adjust temp", "adjust complete", "adjust continue"

        1. "wait for adjust temp": The temperature adjustment of the miner is not completed. 
        2. "adjust complete": The temperature adjustment of the miner is completed, and miner can be power on.
        3. "adjust continue": Miner is adjusting the temperature while waiting to end. 
        
        The value of "complete" is true after the temperature adjustment is complete.
        """
        return self.api.exec_command(self.token, "pre_power_on")



"""
Heavily assisted by and modified from code found from an unknown source on Telegram
with the following credit:
 * @Author: passby
 * @Date: 2020-07-23 00:16:29 
"""
import base64
import binascii
import datetime
import hashlib
import json
import logging
import os
import re
import select
import socket

from base64 import b64encode, b64decode
from Crypto.Cipher import AES
from passlib.hash import md5_crypt

logger = logging.getLogger(__name__)



"""
    Create a WhatsminerAccessToken for each ASIC you want to control and then pass the
    token to the WhatsminerAPI classmethods.

    Basic flow:

    token1 = WhatsminerAccessToken(ip_address="1.2.3.4", admin_password="xxxx")
    token2 = WhatsminerAccessToken(ip_address="1.2.3.5", admin_password="xxxx")

    # Read-only checks
    WhatsminerAPI.get_read_only_info(token1, "status")
    WhatsminerAPI.get_read_only_info(token2, "status")

    # Writeable API
    WhatsminerAPI.exec_command(token1, "power_off", additional_params={"respbefore": "true"})
"""


class WhatsminerAccessToken:
    """ Reusable token to access and/or control a single Whatsminer ASIC.
        Token will renew itself as needed if it expires.
    """
    def __init__(self, ip_address: str, port: int = 4028, admin_password: str = None):
        # Create a read-only access token with just ip_address.
        # Create a read and write access token with ip_address and admin_password
        self.created = datetime.datetime.now()
        self.ip_address = ip_address
        self.port = port
        self._admin_password = admin_password

        if self._admin_password:
            self._initialize_write_access()


    def _initialize_write_access(self):
        """
        Encryption algorithm:
        Ciphertext = aes256(plaintext)，ECB mode
        Encode text = base64(ciphertext)

        (1)api_cmd = token,$sign|api_str    # api_str is API command plaintext
        (2)enc_str = aes256(api_cmd, $key)  # ECB mode
        (3)tran_str = base64(enc_str)

        Final assembly: enc|base64(aes256("token,sign|set_led|auto", $aeskey))
        """
        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect((self.ip_address, self.port))
            s.sendall('{"cmd": "get_token"}'.encode('utf-8'))
            data = recv_all(s, 4000)

        token_info = json.loads(data)["Msg"]
        if token_info == "over max connect":
            raise Exception(data)

        # Make the encrypted key from the admin password and the salt
        pwd = crypt(self._admin_password, "$1$" + token_info["salt"] + '$')
        pwd = pwd.split('$')
        key = pwd[3]

        # Make the aeskey from the key computed above and prep the AES cipher
        aeskey = hashlib.sha256(key.encode()).hexdigest()
        aeskey = binascii.unhexlify(aeskey.encode())
        self.cipher = AES.new(aeskey, AES.MODE_ECB)

        # Make the 'sign' that is passed in as 'token'
        tmp = crypt(pwd[3] + token_info["time"], "$1$" + token_info["newsalt"] + '$')
        tmp = tmp.split('$')
        self.sign = tmp[3]

        self.created = datetime.datetime.now()


    def enable_write_access(self, admin_password: str):
        self._admin_password = admin_password
        self._initialize_write_access()


    def has_write_access(self):
        """ Checks write access and refreshes token, if necessary. """
        if not self._admin_password:
            return False

        if (datetime.datetime.now() - self.created).total_seconds() > 30 * 60:
            # writeable token has expired; reinitialize
            self._initialize_write_access(self._admin_password)

        return True



class WhatsminerAPI:
    """ Stateless classmethod-only read/write API calls. Use a WhatsminerAccessToken
        instance for each ASIC you want to access.
    """

    @classmethod
    def get_read_only_info(self, access_token: WhatsminerAccessToken, cmd: str, additional_params: dict = None):
        """ Send READ-ONLY API command.

            e.g. WhatsminerAPI.get_read_only_info(access_token, cmd="summary")

            Returns: json response
        """
        json_cmd = {"cmd": cmd}
        if additional_params:
            json_cmd.update(additional_params)

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect((access_token.ip_address, access_token.port))
            s.send(json.dumps(json_cmd).encode('utf-8'))
            data = recv_all(s, 4000)

        try:
            return json.loads(data.decode())
        except Exception as e:
            logging.exception("Error calling read-only endpoint")
            try:
                logging.error(data.decode())
            except:
                pass

            raise e


    @classmethod
    def exec_command(self, access_token: WhatsminerAccessToken, cmd: str, additional_params: dict = None):
        """ Send WRITEABLE API command.

            e.g. WhatsminerAPI.exec_command(access_token, cmd="power_off", additional_params={"respbefore": "true"})

            Returns: json response
        """
        if not access_token.has_write_access():
            raise Exception("access_token must have write access")

        # Assemble the plaintext json
        json_cmd = {"cmd": cmd, "token": access_token.sign}
        if additional_params:
            json_cmd.update(additional_params)
        api_cmd = json.dumps(json_cmd)

        # Encrypt it and assemble the transport json
        enc_str = str(
            base64.encodebytes(
                access_token.cipher.encrypt(add_to_16(api_cmd))),
                encoding='utf8'
            ).replace('\n', '')
        data_enc = {'enc': 1}    # transmit w/ "enc" to signal that it's encrypted
        data_enc['data'] = enc_str
        api_packet_str = json.dumps(data_enc)

        with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
            s.connect((access_token.ip_address, access_token.port))
            s.send(api_packet_str.encode())
            data = recv_all(s, 4000)

        try:
            json_response = json.loads(data.decode())
            if "STATUS" in json_response and json_response["STATUS"] == "E":
                logger.error(json_response["Msg"])
                raise Exception(api_cmd + "\n" + json_response["Msg"])

            resp_ciphertext = b64decode(json.loads(data.decode())["enc"])
            resp_plaintext = access_token.cipher.decrypt(resp_ciphertext).decode().split("\x00")[0]
            resp = json.loads(resp_plaintext)
        except Exception as e:
            logger.exception("Error decoding encrypted response")
            try:
                logger.error(data.decode())
            except:
                pass
            raise e

        return resp



# ================================ misc helpers ================================
def crypt(word, salt):
    standard_salt = re.compile('\s*\$(\d+)\$([\w\./]*)\$')
    match = standard_salt.match(salt)
    if not match:
        raise ValueError("salt format is not correct")
    extra_str = match.group(2)
    result = md5_crypt.hash(word, salt=extra_str)
    return result


# Adapted from: https://stackoverflow.com/a/17668009
def recv_all(sock, n):
    # Helper function to recv n bytes
    BUFF_SIZE = 4096 # 4 KiB
    data = b''
    while True:
        part = sock.recv(BUFF_SIZE)
        data += part
        if len(part) < BUFF_SIZE:
            # either 0 or end of data
            break
    return data


def add_to_16(s):
    while len(s) % 16 != 0:
        s += '\0'
    return str.encode(s)  # return bytes



from pywhatsminer.core.api import WhatsminerAPI, WhatsminerAccessToken
from typing import Any


class Config:
    def __init__(self, client):
        self.client = client
        self.api: WhatsminerAPI = client.api
        self.token: WhatsminerAccessToken = client._access_token
        
        
    def open_ssh(self) -> Any:
        """
        This operation opens miner's SSH server.
        """
        return self.api.exec_command(self.token, "ssh_open")
    

    def close_ssh(self) -> Any:
        """
        This operation closes miner's SSH server.
        """
        return self.api.exec_command(self.token, "ssh_close")
    
    
    def manage_led(self, color: str | None = None, period: int | None = None, duration: int | None = None, start: int | None = None) -> Any:
        """
        This operation manages miner's LED. If no fields are provided, it sets to auto control mode.
        
        Params:
            - color: str = None - LED color. Possible fields: "green", "red"
            - period: int = None - LED flash cycle (in ms)
            - duration: int = None - LED on time in cycle (in ms)
            - start: int = None - LED time offset (in ms)
        """
        
        assert color in ["green", "red"]
        
        if not any([color, period, duration, start]):
            return self.api.exec_command(self.token, "set_led", {"param":"auto"})
        else:
            return self.api.exec_command(self.token, "set_led", {"param":color, "period":period, "duration":duration, "start":start})
        
        
    def change_password(self, old: str, new: str) -> Any:
        """
        This operation changes miner's admin password. Requires old password.
        
        Params:
            - old: str - old password
            - new: str - new password
        """
        return self.api.exec_command(self.token, "update_pwd", {"old":old, "new":new})
    
    
    def set_target_frequency(self, frequency: int) -> Any:
        """
        This operation sets miner's target frequency.
        
        Params:
            - freq: int - target frequency in percent. Range: -100 to 100.
        """
        assert frequency in range(-100, 101)
        
        return self.api.exec_command(self.token, "set_target_freq", {"percent":frequency})
    
    
    def enable_fastboot(self) -> Any:
        """
        This operation enables btminer fastboot. After setting, the next restart of the miner takes effect.
        """
        return self.api.exec_command(self.token, "enable_btminer_fast_boot")
    
    
    def disable_fastboot(self) -> Any:
        """
        This operation disables btminer fastboot. After setting, the next restart of the miner takes effect.
        """
        return self.api.exec_command(self.token, "disable_btminer_fast_boot")
    
    
    def enable_web_pools(self) -> Any:
        """
        This operation allows configuration of pools on web pages with immediate effect.

        """
        return self.api.exec_command(self.token, "enable_web_pools")
    
    
    def disable_web_pools(self) -> Any:
        """
        This operation turns off the configure pools feature on the web page with immediate effect.
        """
        return self.api.exec_command(self.token, "disable_web_pools")
    
    
    def set_hostname(self, hostname: str) -> Any:
        """
        This operation sets miner's hostname, does not take effect until the network is restarted.
        
        Params:
            - hostname: str - new hostname
        """
        return self.api.exec_command(self.token, "set_hostname", {"hostname":hostname})
    
    
    def set_time_zone(self, timezone: str) -> Any:
        """
        This operation sets miner's time zone, does not take effect until the network is restarted.
        
        Params:
            - timezone: str - new timezone (eg. "CST-8")
            - zonename: str - new timezone (eg. "Asia/Shanghai")
        """
        return self.api.exec_command(self.token, "set_timezone", {"timezone":timezone})
    
    
    def load_log(self, ip: str, port: int | str, proto: str) -> Any:
        """
        This operation configures the rsyslog log server. 
        
        Params:
            - ip: str - ip address
            - port: int | str - port
            - proto: str - protocol. Possible fields: "tcp", "udp"
        """
        assert proto in ["tcp", "udp"]
        
        return self.api.exec_command(self.token, "load_log", {"ip":ip, "port":port, "proto":proto})
    
    
    def set_temp_offset(self, offset: int) -> Any:
        """
        This operation sets offset of miner hash board target temperature. 
        
        Params:
            - offset: int - temperature offset in degrees. Range: -30 to 0.
        """
        assert offset in range(-30, 1)
        
        return self.api.exec_command(self.token, "set_temp_offset", {"temp_offset":offset})
    

    def adjust_upfreq_speed(self, speed: int) -> Any:
        """
        This operation sets the upfreq speed, 0 is the normal speed, 9 is the fastest speed. 
        The faster the speed, the greater the final hash rate and power deviation, and the stability may have a certain impact. 
        Fast boot mode cannot be used at the same time.
        
        Params:
            - speed: int - new upfreq speed. Range: 0 to 9.
        """
        assert speed in range(0, 10)
        
        return self.api.exec_command(self.token, "adjust_upfreq_speed", {"upfreq_speed":speed})
    
    
    def set_poweroff_cool(self, poweroff_cool: bool) -> Any:
        """
        This operation sets whether to cool machine when stopping mining.
        
        Params:
            - poweroff_cool: bool - whether to cool machine when stopping mining
        """
        return self.api.exec_command(self.token, "set_poweroff_cool", {"poweroff_cool":1 if poweroff_cool else 0})
    
    
    def set_fan_zero_speed(self, fan_zero_speed: bool) -> Any:
        """
        This operation sets whether the fan speed supports the lowest 0 speed.
        
        Params:
            - fan_zero_speed: bool - whether the fan speed supports the lowest 0 speed
        """
        return self.api.exec_command(self.token, "set_fan_zero_speed", {"fan_zero_speed":1 if fan_zero_speed else 0})


from setuptools import setup, find_packages


setup(
  name = 'pywhatsminer',
  version = '0.1.8',
  license='MIT',
  description = 'Unofficial Python based Api-Client (Wrapper) for MicroBT Whatsminer ASIC`s',
  long_description="See full README at https://github.com/DAAMCS/PyWhatsminer",
  author = 'DAAMCS | Forked from Satoshi Anonymoto',
  author_email = 'me@daamcs.ru',
  url = 'https://github.com/DAAMCS/PyWhatsminer',
  download_url = 'https://github.com/DAAMCS/PyWhatsminer',
  keywords = ['whatsminer', 'microbt', 'api', 'btc', 'asic'],
  install_requires=[
          'passlib',
          'pycryptodome',
          'dataclasses-json',
          'Deprecated'
      ],
  classifiers=[

    'Development Status :: 4 - Beta',
    'Intended Audience :: Developers',
    'Topic :: Software Development :: Build Tools',
    'License :: OSI Approved :: MIT License',
    'Programming Language :: Python :: 3',
  ],
  packages=find_packages(),
  include_package_data=True,
)




from pywhatsminer.core.api import WhatsminerAPI, WhatsminerAccessToken
from pywhatsminer.core.models import Summary, Status, Api, Dev, Devs, DevDetail, DevDetails, Pool, Pools, Info, ErrorCode, Errors
from pywhatsminer.core.utils import process_response

from typing import Any

from deprecated import deprecated


class System:
    """
    This class provides methods for miner's system.
    """
    def __init__(self, client):
        self.client = client
        self.api: WhatsminerAPI = client.api
        self.token: WhatsminerAccessToken = client._access_token
        
    
    def reboot(self) -> Any:
        """
        This operation simply reboots miner's system.
        """
        return self.api.exec_command(self.token, "reboot")


    def reset(self) -> Any:
        """
        This operation resets the miner to factory settings.
        """
        return self.api.exec_command(self.token, "factory_reset")
    
    
    # TODO: Check status model
    def get_status(self) -> Status:
        """
        This method returns miner's status.
        
        WARNING: Response model can be incorrect. Will be fixed in the future.
        """
        data = process_response(self.api.exec_command(self.token, "status"))

        return Status(*data['Msg'].values())
    
    
    def get_summary(self) -> Summary:
        """
        This method returns miner's summary.
        """
        data = process_response(self.api.exec_command(self.token, "summary"))
        summary = process_response(data['SUMMARY'][0])
        
        return Summary(*summary.values())
    
    
    def get_api_version(self) -> Api:
        """
        This method returns miner's API version.
        """
        data = process_response(self.api.exec_command(self.token, "get_version"))
    
        return Api(*data['Msg'].values())
    
    
    def get_dev_details(self) -> DevDetails:
        """
        This method returns miner's device details.
        """
        data = self.api.exec_command(self.token, "devdetails")
        devdetails = DevDetails(details=[])
        
        for devdetail in data['DEVDETAILS']:
            devdetail = process_response(devdetail)
            devdetail.pop('ID')
            devdetails.details.append(DevDetail(*devdetail.values()))
    
        return devdetails
    
    
    def get_devs(self) -> Devs:
        """
        This method returns information for each hash board.
        """
        data = self.api.exec_command(self.token, "devs")
        devs = Devs(devs=[])
    
        for dev in data['DEVS']:
            dev = process_response(dev)
            devs.devs.append(Dev(*dev.values()))
    
        return devs
    
    
    def get_pools(self) -> Pools:
        """
        This method returns pool miner information.
        """
        data = self.api.exec_command(self.token, "pools")
        pools = Pools(pools=[])
        
        for pool in data['POOLS']:
            pool = process_response(pool)
            pools.pools.append(Pool(*pool.values()))
            
        return pools    
    
    
    def get_miner_info(self, *fields: str) -> Any:
        """
        This method returns miner's information.
        Params:
            - *fields: str - you can select the fields that you want to return, returns all if not specified.
        
        WARNING: looks like fields are not working, in any case api returns all fields in response.
        """
        data = self.api.exec_command(self.token, "get_miner_info", {'info':f"{",".join(fields)}"})
        info = Info(**data['Msg'])
        
        return info
    
    
    # TODO: Implement this
    def get_error_codes(self) -> Any:
        """
        This method returns miner's error codes.
        
        WARNING: This method actually don't parse response and gives raw json in response. Will be implemented in the future.
        """
        data = self.api.exec_command(self.token, "get_error_code")
        
        return data

    
    # TODO: Implement this
    @deprecated(reason="Not work. Will be implemented in the future.")
    def get_token(self) -> Any:
        """
        This method returns miner's access token.
        
        WARNING: This method actually don't work. Will be implemented in the future.
        """
        return self.api.exec_command(self.token, "get_token")
    


from dataclasses import dataclass
from dataclasses_json import dataclass_json

from datetime import datetime, timedelta


@dataclass_json
@dataclass
class Summary:
    elapsed: int
    mhs_avg: float
    mhs_5s: float
    mhs_1m: float
    mhs_5m: float
    mhs_15m: float
    hs_rate: float
    accepted: int
    rejected: int
    total_mh: float
    temperature: float
    avg_frequency: int
    fan_speed_in: int
    fan_speed_out: int
    power_watts: int
    power_rate: float
    pool_rejected_percent: float
    pool_stale_percent: float
    uptime: timedelta
    target_frequency: int
    target_mhs: float
    env_temp: float
    power_mode: str
    factory_ghs: int
    power_limit: int
    chip_temp_min: float
    chip_temp_max: float
    chip_temp_avg: float
    btminer_fast_boot_enabled: str
    upfreq_complete: bool
    

# TODO: Validate models, maybe incorrect. Needs some tests.
@dataclass_json
@dataclass
class Status:
    miner_off: bool
    miner_off_reason: str
    miner_off_time: int
    firmware_version: str
    power_mode: str
    power_limit_set: bool
    hash_percent: int


@dataclass_json
@dataclass
class PSU:
    name: str
    hardware_version: str
    software_version: str 
    model: str
    enabled: bool
    iin: int
    vin: int
    pin: int
    fan_speed: int
    serial_no: str
    vendor: int
    temperature: float
    
    
@dataclass_json
@dataclass
class Api:
    version: str
    firmware_version: str
    platform: str
    chip: str
    
    
@dataclass_json
@dataclass
class DevDetail:
    id: int
    name: str
    driver: str
    kernel: str
    model: str
    
    
@dataclass_json
@dataclass
class DevDetails:
    details: list[DevDetail]
    
    
@dataclass_json
@dataclass
class Dev:
    asc: int
    slot: int
    enabled: bool
    is_alive: bool
    temperature: float
    chip_frequency: int
    mhs_av: float
    mhs_5s: float
    mhs_1m: float
    mhs_5m: float
    mhs_15m: float
    hs_rate: float
    factory_ghs: int
    upfreq_complete: bool
    effective_chips: int
    pcb_serial_no: str
    chip_data: str
    chip_temp_min: float
    chip_temp_max: float
    chip_temp_avg: float
    chip_vol_diff: int
    

@dataclass_json
@dataclass
class Devs:
    devs: list[Dev]
    

@dataclass_json
@dataclass
class Pool:
    id: int
    url: str
    is_alive: bool
    priority: int
    quota: int
    getworks: int
    accepted: int
    rejected: int
    works: int
    discarded: int
    stale: int
    get_failures: int
    remote_failures: int
    user: str
    last_share_time: datetime
    stratum_is_active: bool
    stratum_difficulty: float
    pool_rejected_percent: float
    pool_stale_percent: float
    bad_works: int
    current_block_height: int
    current_block_version: int
    to_remove: bool
    

@dataclass_json
@dataclass
class Pools:
    pools: list[Pool]
    
    
@dataclass_json
@dataclass
class Info:
    ntp: list[str] | None = None
    ip: str | None = None
    proto: str | None = None
    netmask: str | None = None
    gateway: str | None = None
    dns: str | None = None
    hostname: str | None = None
    mac: str | None = None
    ledstat: str | None = None
    minersn: str | None = None
    powersn: str | None = None
    
    
@dataclass_json
@dataclass
class ErrorCode:
    code: int
    appeared_at: datetime


@dataclass_json
@dataclass
class Errors:
    codes = list[ErrorCode]